; Some basic, easy theorems about the definitions in ordinal-definitions.lisp
(in-package "ACL2")

(include-book "ordinal-total-order")
(local (disable-forcing))

; theorems about natp, posp

(defthm natp-cr
  (equal (natp x)
         (and (integerp x)
              (<= 0 x)))
  :rule-classes :compound-recognizer)

(defthm posp-cr
  (equal (posp x)
         (and (integerp x)
              (< 0 x)))
  :rule-classes :compound-recognizer)

(defthm natp-op-type
  (implies (natp x)
           (op x))
  :rule-classes :forward-chaining)

(defthm posp-op-type
  (implies (posp x)
           (op x))
  :rule-classes :forward-chaining)

(defthm |x < y  =>  0 < -x+y|
  (implies (and (integerp x)
                (integerp y)
                (< x y))
           (posp (+ (- x) y)))
  :rule-classes ((:type-prescription)
                 (:forward-chaining :trigger-terms ((+ (- x) y)))))

(defthm |x < y  =>  0 < y-x|
  (implies (and (integerp x)
                (integerp y)
                (< x y))
           (posp (+ y (- x))))
  :rule-classes ((:type-prescription)
                 (:forward-chaining :trigger-terms ((+ y (- x))))))

(defthm |x <= y  =>  0 <= y-x|
  (implies (and (integerp x)
                (integerp y)
                (<= x y))
           (natp (+ y (- x))))
  :rule-classes ((:type-prescription)
                 (:forward-chaining :trigger-terms ((+ y (- x))))))

(defthm posp-+
  (implies (and (posp a)
		(posp b))
	   (posp (+ a b))))

(defthm posp-*
 (implies (and (posp a)
	       (posp b))
	  (posp (* a b))))

(defthm natp-expt
  (implies (and (natp a)
		(natp b))
	   (natp (expt a b))))

(defthm posp-op
  (implies (and (not (equal x 0))
                (op x)
                (not (consp x)))
           (posp x))
  :rule-classes :forward-chaining)

(defthm posp->-0
  (implies (posp a)
	   (< 0 a)))

(defthm natp->=0
  (implies (natp a)
	   (<= 0 a)))

(in-theory (disable natp posp))


; theorems involving first-exp and first-coef

(defthm cons-cons-equal
  (implies (and (consp a)
		(op a))
	   (equal (cons (cons (first-exp a) (first-coef a))
                        (cdr a))
		  a)))

(defthm fe-fc-cdr-equal
  (implies (and (op a)
		(op b))
	   (equal (and (equal (first-exp a)
			      (first-exp b))
		       (equal (first-coef a)
			      (first-coef b))
		       (equal (cdr a)
			      (cdr b)))
		  (equal a b)))
  :rule-classes 
  ((:forward-chaining :trigger-terms ((equal (first-exp a)
                                             (first-exp b))
                                      (equal (first-coef a)
                                             (first-coef b))
                                      (equal (cdr a)
                                             (cdr b))))))

;theorems about first-exp

(defthm first-exp-equal
  (implies (not (equal (first-exp a) (first-exp b)))
	   (not (equal a b)))
  :rule-classes :forward-chaining)

(defthm op-first-exp
  (implies (op a)
	   (op (first-exp a)))
  :rule-classes ((:rewrite)
		 (:forward-chaining)))

(defthm |(first-exp a) < a|
 (implies (and (op a)
	       (not (equal a 0)))
	  (o< (first-exp a) a)))

(defthm first-exp-<
 (implies (and (op a)
	       (op b)
	       (o< (first-exp a)
		   (first-exp b)))
	  (o< a b)))

(defthm first-exp-<=
  (implies (op b)
           (<= 0 (first-exp b)))
  :rule-classes :linear)

(defthm <-first-exp-<=
  (implies (and (op a)
		(op b)
		(o< a b)
		(not (equal (first-exp a)
			    (first-exp b))))
	   (o< (first-exp a)
	       (first-exp b)))
  :rule-classes 
  ((:forward-chaining 
    :trigger-terms ((equal (first-exp a)
                           (first-exp b))
                    (o< (first-exp a)
                        (first-exp b))
                    (o< a b)))))

(defthm first-exp-natp
  (implies (and (equal (first-exp a)
		       0)
                (op a))
	   (natp a))
  :rule-classes :forward-chaining)

(defthm first-exp-<-consp-1
  (implies (and (not (o< (first-exp b)
			 (first-exp a)))
                (consp a)
		(op a))
	   (consp b))
  :rule-classes :forward-chaining)

(defthm first-exp-<-consp-2
  (implies (and (o< (first-exp a)
		    (first-exp b))
		(op a)
		(op b))
	   (consp b))
  :rule-classes :forward-chaining)

(defthm first-exp-equal-atom
  (implies (and (equal (first-exp a)
		       (first-exp b))
		(atom a)
		(op b))
	   (atom b))
  :rule-classes :forward-chaining)

(defthm first-exp-def-consp
  (equal (first-exp (cons (cons a b) c))
         a))

(defthm first-exp-def-atom
  (implies (atom a)
	   (equal (first-exp a)
		  0)))

(in-theory (disable first-exp))

;theorems about first-coef

(defthm first-coef-equal
  (implies (not (equal (first-coef a)
		       (first-coef b)))
	   (not (equal a b)))
  :rule-classes :forward-chaining)

(defthm op-first-coef
 (implies (op a)
	  (op (first-coef a))))

(defthm natp-first-coef
  (implies (op a)
	   (natp (first-coef a)))
  :rule-classes :forward-chaining)

(defthm posp-first-coef
  (implies (and (op a)
                (not (equal a 0)))
	   (posp (first-coef a)))
  :rule-classes ((:rewrite)
		 (:forward-chaining)))

(defthm first-coef-def-consp
  (equal (first-coef (cons (cons a b) c))
         b))

(defthm first-coef-def-atom
  (implies (atom a)
	   (equal (first-coef a) a)))

(in-theory (disable first-coef))

;theorems about op

(defthm op-consp-car
  (implies (and (consp a)
		(op a))
	   (consp (car a)))
  :rule-classes ((:rewrite)
		 (:forward-chaining)))

(defthm op-fe-neq-0
  (implies (and (consp a)
		(op a))
	   (not (equal (first-exp a) 0)))
  :rule-classes ((:rewrite)
		 (:forward-chaining)))

(defthm op-op-cdr
  (implies (and (consp a)
		(op a))
	   (op (cdr a)))
  :rule-classes ((:rewrite)
		 (:forward-chaining)))

(defthm op-<-fe-fe-cdr
  (implies (and (consp a)
		(op a))
	   (o< (first-exp (cdr a))
	       (first-exp a)))
  :rule-classes ((:rewrite)
		 (:forward-chaining)))

(defthm op-fc
  (implies (op (cons a b))
           (op b))
  :rule-classes :forward-chaining)

(defthm op-def-consp
  (implies (and (op a)
                (not (equal a 0))
                (posp b)
                (op c)
                (o< (first-exp c) a))
           (op (cons (cons a b) c))))

(defthm op-def-atom
  (implies (atom a)
	   (equal (op a)
		  (natp a))))

(in-theory (disable op))
