;  This files contains some theorems about limit ordinals.  The
;  purpose is to show how to define and prove theorems about
;  limit ordinals, but the theory of limit ordinals is not
;  developed fully.

(in-package "ACL2")

(include-book "ordinal-exponentiation")

(defthm limitp-op
  (implies (limitp a)
           (op a))
  :rule-classes :forward-chaining)

(defthm limitp-consp
  (implies (limitp a)
           (consp a))
  :rule-classes :forward-chaining)

(defthm limitp-cdr-natp
  (implies (and (limitp a)
                (natp (cdr a)))
           (equal (cdr a) 0))
  :rule-classes :forward-chaining)

(defthm limitp-cdr
  (implies (limitp a)
           (or (equal (cdr a) 0)
               (limitp (cdr a))))
  :rule-classes :forward-chaining)

(defthm |limit.b  =>  limit(a+b)|
  (implies (and (limitp b)
                (op a))
           (limitp (o+ a b)))
  :hints (("goal" :induct (o+ a b))))

(defthm limit-+
  (implies (op a)
           (not (limitp (o+ a 1))))
  :hints (("goal" :induct (o+ a 1))))

(in-theory (disable limitp))

(defthm |limitp.b  =>  a<b   <=>  a+1 < b|
  (implies (and (limitp b)
                (op a))
           (equal (o< (o+ a 1) b)
                  (o< a b)))
  :hints (("goal" :in-theory (disable o+))))

(defthm |limitp.b  =>   c<b  <=>  a+c+1 < a+b|
  (implies (and (limitp b)
                (op a)
                (op c))
           (equal (o< (o+ a c 1) (o+ a b))
                  (o< c b)))
  :hints (("goal" :in-theory (disable o+))))

(defthm atom-fc-op
  (implies (and (equal (first-coef a) 0)
                (op a))
           (atom a))
  :rule-classes :forward-chaining)

(defthm |limitp.b /\  a < w^b   =>   a < w^[fe.a +1]  /\  [fe.a + 1] < w^b|
  (implies (and (limitp b)
                (op a)
                (o< a (o^ (omega) b)))
           (and (o< a 
                    (o^ (omega) (o+ (first-exp a) 1)))
                (o< (o^ (omega) (o+ (first-exp a) 1))
                    (o^ (omega) b))))
  :rule-classes ((:rewrite :match-free :all)))

(defthm o*-fe-equal
  (implies (and (not (equal a 0))
		(not (equal b 0))
		(equal (first-exp a)
		       (first-exp b))
		(limitp c)
		(op a)
		(op b))
	   (equal (o* a c)
		  (o* b c)))
  :rule-classes ((:forward-chaining :trigger-terms ((o* a c) (o* b c)))))
