;definitions and theorems about ordinal arithmetic
(in-package "ACL2")
(include-book "basic-definitions-thms")
(include-book "top-with-meta")
(local (disable-forcing))

(defthm o+-atomic
  (implies (and (not (consp a))
		(not (consp b)))
	   (not (consp (o+ a b))))
  :rule-classes ((:type-prescription)
                 (:forward-chaining  :trigger-terms ((o+ a b)))))

(defthm o+-atomic-2
  (implies (and (posp a)
		(natp b))
	   (posp (o+ a b)))
  :rule-classes ((:type-prescription)
                 (:forward-chaining :trigger-terms ((o+ a b)))))

(defthm consp-o+
  (implies (and (consp a)
                (op a))
	   (consp (o+ a b)))
  :rule-classes ((:forward-chaining :trigger-terms ((o+ a b)))
                 (:rewrite)))

(defthm consp-o+-2
  (implies (and (op b)
		(consp b))
           (consp (o+ a b)))
  :rule-classes ((:forward-chaining  :trigger-terms ((o+ a b)))
                 (:rewrite)))

(defthm consp-o+-3
  (implies (and (consp a)
		(consp b))
	   (consp (o+ a b)))
  :rule-classes ((:forward-chaining  :trigger-terms ((o+ a b)))
                 (:rewrite)))

(defthm o+-first-exp-1
  (implies (o< (first-exp a)
	       (first-exp b))
	   (equal (first-exp (o+ a b))
		  (first-exp b))))

(defthm o+-first-exp-2
  (implies (and (op b)
		(not (o< (first-exp a)
			 (first-exp b))))
	   (equal (first-exp (o+ a b))
		  (first-exp a))))

(defthm o+-first-coef-1
 (implies (o< (first-exp a)
	      (first-exp b))
	  (equal (first-coef (o+ a b))
		 (first-coef b))))

(defthm o+-first-coef-2
  (implies (and (op b)
		(o< (first-exp b)
		    (first-exp a)))
	   (equal (first-coef (o+ a b))
		  (first-coef a)))
  :hints (("goal"
	   :expand (o+ a b))))

(defthm o+-first-coef-3
  (implies (and (equal (first-exp a)
		       (first-exp b))
                (op b))
	   (equal (first-coef (o+ a b))
		  (+ (first-coef a) (first-coef b)))))

(defthm fe-o+-<
 (implies (and (op a)
	       (op c))
	  (equal (o< (first-exp (o+ b c))
                     (first-exp a))
                 (and (o< (first-exp b)
                          (first-exp a))
                      (o< (first-exp c)
                          (first-exp a))))))

(defthm op-o+
  (implies (and (op a)
		(op b))
	   (op (o+ a b))))

(defthm |a+0 = a|
  (implies (op a)
           (equal (o+ a 0) a)))

(defthm |0+a = a|
  (implies (op a)
           (equal (o+ 0 a) a)))

; Notice that <= is <= on numbers.
(defthm |0 <= a+b|
  (implies (and (op a)
                (op b))
           (<= 0 (o+ a b))))

(defthm |a < a+1|
  (implies (op a)
           (o< a (o+ a 1))))

(local
 (defun add-ind-1 (a b)
   (if (consp a)
       (add-ind-1 (cdr a) b)
     (list a b))))

; The next theorem seems obvious, but is needed to augment ACL2's
; ability to reason about equality.
(defthm cons-cons-equal-2
 (implies (and (consp a)
	       (op a)
	       (op b)
	       (equal (first-exp a)
		      (first-exp b))
	       (equal (first-coef a)
		      (first-coef b)))
	  (equal (cons (cons (first-exp a)
			     (first-coef a)) 
		       (cdr b))
		 b))
 :hints (("goal"
	  :use ((:instance cons-cons-equal
			   (a b))))))

(defthm |a < b  <=>  a+1 <= b|
  (implies (and (op a)
                (op b))
           (equal (o< (o+ a 1) b)
                  (and (o< a b)
                       (not (equal (o+ a 1) b))))))

(defthm |a < b+1  <=>  a <= b|
  (implies (and (op a)
                (op b))
           (equal (o< a (o+ b 1))
		  (or (o< a b)
		      (equal a b)))))

(defthm |a <= b+a|
  (implies (and (op a)
		(op b))
           (equal (o< a (o+ b a))
                  (not (equal (o+ b a) a)))))

;without the hint, this proof requires induction!
(defthm |~(b+a < a)|
  (implies (and (op a)
		(op b))
	   (not (o< (o+ b a) a)))
  :hints (("goal" :use ((:instance |a <= b+a|)))))

(defthm |a < a+b  <=>  ~(0 = b)| ;|a <= a+b|
  (implies (and (op a)
		(op b))
           (equal (o< a (o+ a b))
                  (not (equal b 0)))))

(defthm |x+y = 0  <=>  x=0 /\ y=0|
  (implies (and (op x)
                (op y))
           (equal (equal 0 (o+ x y))
                  (and (equal x 0)
                       (equal y 0)))))

(local
 (defun o+-induct (a b c)
   (if (and (op a)
	    (op b)
	    (op c)
	    (o< (first-exp b) (first-exp a))
	    (o< (first-exp c) (first-exp a)))
       (o+-induct (cdr a) b c)
     (list a b c))))

(defthm o+-fe-smash
  (implies (o< (first-exp a)
	       (first-exp b))
	   (equal (o+ a b) b)))

(defthm fe-<-o+
  (implies (and (op b)
		(op c))
	   (equal (o< (first-exp a)
		      (first-exp (o+ b c)))
		  (or (o< (first-exp a)
			  (first-exp b))
		      (o< (first-exp a)
			  (first-exp c)))))
  :hints (("goal"
	   :cases ((o< (first-exp b)
		       (first-exp c))))))

(encapsulate 

; This encapsulate is here just so that we do not have to give a
; subgoal hint to |(a+b)+c = a+(b+c)|.

 nil

 (local
  (defthm |(a+b)+c = a+(b+c) helper|
    (implies (and (not (o< (first-exp c) (first-exp a)))
                  (op a)
                  (op b)
                  (op c))
             (equal (o+ (o+ a b) c)
                    (o+ a (o+ b c))))
    :hints (("goal"
             :cases ((o< (first-exp b) (first-exp c))
                     (o< (first-exp c) (first-exp b)))))))

 (defthm |(a+b)+c = a+(b+c)|
   (implies (and (op a)
                 (op b)
                 (op c))
            (equal (o+ (o+ a b) c)
                   (o+ a (o+ b c))))
   :hints (("goal"
            :induct (o+-induct a b c)))))

(defthm |a < b  <=>  c+a < c+b|
  (implies (and (op a)
		(op b)
                (op c))
           (equal (o< (o+ c a)
                      (o+ c b))
                  (o< a b))))

(defthm |a-a = 0|
  (equal (o- a a) 0))

(defthm op-o-
  (implies (and (op a)
                (op b))
           (op (o- a b))))

(defthm |a-b <= a|
  (implies (and (op a)
                (op b))
           (equal (o< (o- a b) a)
                  (not (equal (o- a b) a))))
  :hints (("goal" 
           :induct (o- a b))))

(defthm o--first-exp
  (implies (and (op a)
		(op b))
	   (not (o< (first-exp a) 
		    (first-exp (o- a b))))))

(defthm cons-cons-equal-3
  (implies (and (consp a)
		(op a)
		(op b)
		(equal (first-exp a)
		       (first-exp b)))
	   (equal (cons (cons (first-exp a)
			      (first-coef b))
			(cdr b))
		  b))
  :hints (("goal"
	   :use ((:instance cons-cons-equal
			    (a b))))))

; The hints are not really needed, but they speed up the book. 
(defthm fe-o--<
 (implies (and (op a)
	       (op b)
	       (op c)
	       (o< (first-exp b) (first-exp a)))
	  (o< (first-exp (o- b c))
	      (first-exp a)))
 :hints (("goal"
	  :in-theory 
	  (disable |a <= b /\ b < c  =>  a < c|)
	  :use ((:instance |a <= b /\ b < c  =>  a < c|
			   (a (first-exp (o- b 
					     c)))
			   (b (first-exp b))
			   (c (first-exp a)))))))

(encapsulate
 ()

 (local
  (defthm lemma-1
    (implies (and (consp a)
		  (equal c (first-exp a))
		  (op a)
		  (op b))
	     (o< (first-exp (o- (cdr a) b)) c))
  :hints (("goal"
	   :use ((:instance fe-o--<
			    (b (cdr a))
			    (c b))))))) 
 (local
  (defun o-induct (x y)
    (if (and (consp x)
	     (consp y)
	     (equal (first-exp x) (first-exp y))
	     (equal (first-coef x) (first-coef y)))
	(list (o-induct (cdr x) (cdr y)))
      (list x y))))

 (defthm |a < b  =>  a+(b-a) = b|
   (implies (and (op a)
		 (op b)
		 (o< a b))
	    (equal (o+ a (o- b a))
		   b))
   :hints (("goal" 
	    :induct (o-induct b a)))))
	   
#|

The purpose of the next two local events is just to speed up the proof of
theorem |a < b => a+c <= b+c| (which goes through without the lemmas and
without any hints but takes 123 seconds, whereas the time taken by the lemmas
and the theorem is now less than 3 seconds.  It probably was not worth the effort
involved, but I wanted to see why ACL2 couldn't get it.  It is an interesting
example. 

|#
(encapsulate
 ()
 (local
  (defthm cnf-lemma-1
    (implies (and (op b)
		  (op a)
		  (op c)
		  (o< a (o+ a (o- b a)))
		  (not (equal (o+ a c)
			      (o+ (o+ a (o- b a)) c))))
	     (o< (o+ a c)
		 (o+ (o+ a (o- b a)) c)))
    :hints (("goal"
	     :use ((:instance |a < b  <=>  c+a < c+b| 
			      (a c)
			      (c a)
			      (b (o+ (o- b a) c)))
		   (:instance |a <= b+a| 
			      (a c) 
			      (b (o- b a))))))))
 
 (local
  (defthm cnf-lemma-2
    (implies (and (equal (o+ a (o- b a)) b)
		  (op a)
		  (op (o+ a (o- b a)))
		  (op c)
		  (o< a (o+ a (o- b a)))
		  (not (equal (o+ a c)
			      (o+ (o+ a (o- b a)) c))))
	     (o< (o+ a c)
		 (o+ (o+ a (o- b a)) c)))
    :otf-flg t
    :hints (("goal"
	     :do-not-induct t
	     :use ((:instance |a < b  <=>  c+a < c+b| 
			      (a c)
			      (c a)
			      (b (o+ (o- b a) c)))
		   (:instance cnf-lemma-1))))))

 (defthm |a < b  =>  a+c <= b+c|
   (implies (and (op a)
		 (op b)
		 (op c)
		 (o< a b))
	    (equal (o< (o+ a c) 
                       (o+ b c))
                   (not (equal (o+ a c) 
                               (o+ b c)))))
   :hints (("goal"
	    :use ((:instance |a < b  =>  a+(b-a) = b|)
		  (:instance cnf-lemma-2))
	    :in-theory (disable |a < b  =>  a+(b-a) = b|)))))

(defthm |a+c = b  =>  b-a = c|
  (implies (and (equal (o+ a c) b)
                (op a)
                (op b)
                (op c))
           (equal (o- b a) c))
  :rule-classes :forward-chaining)

(defthm |(a+b)-a = b|
  (implies (and (op a)
                (op b))
           (equal (o- (o+ a b) a)
                  b)))

(defthm |a+b = a+c  <=>  b=c|
  (implies (and (op a)
                (op b)
                (op c))
           (equal (equal (o+ a b)
                         (o+ a c))
                  (equal b c))))

(defthm o--0
 (implies (o< a b)
	  (equal (o- a b) 0)))


(defthm |a <= b  <=>  b-a = 0|
 (implies (and (op a)
	       (op b))
          (equal (equal (o- b a) 0)
                 (not (o< a b)))))

(defthm o+-atom-def
  (implies (and (atom a)
		(atom b))
	   (equal (o+ a b) (+ a b))))
  	
(defthm o+-cdr-1
  (implies (and (op a)
		(op b)
		(o< (first-exp b) 
		    (first-exp a)))
	   (equal (cdr (o+ a b)) (o+ (cdr a) b))))

(defthm o+-cdr-2
  (implies (and (op a)
		(op b)
		(not (o< (first-exp b)
			 (first-exp a))))
	   (equal (cdr (o+ a b)) (cdr b))))

(defthm o+-consp-def-1
  (equal (o+ (cons (cons x a2) a3)
             (cons (cons x b2) b3))
         (cons (cons x (+ a2 b2)) b3)))

(defthm o+-cdr-equal
  (implies (and (consp a)
		(op a))
	   (equal (o+ (cons (cons (first-exp a) (first-coef a))
			    0)
		      (cdr a))
		  a)))
