(in-package "ACL2")
(include-book "ordinal-multiplication")
(local (disable-forcing))
(local (in-theory (disable o- o+ o*)))

(defthm |a^0 = 1|
  (equal (o^ a 0) 1))

(defthm |a^1 = a|
  (implies (op a)
	   (equal (o^ a 1) a)))

(defthm |~(0 = a)  <=>  0^a = 0|
  (implies (op a)
	   (equal (equal (o^ 0 a) 0)
		  (not (equal a 0)))))

(defthm |0 < a  =>  0^a = 0 :a|
  (implies (posp a)
	   (equal (o^ 0 a) 0))
  :rule-classes ((:type-prescription)))

(defthm |0 < a  =>  0^a = 0 :b|
  (implies (consp a)
	   (equal (o^ 0 a) 0))
  :rule-classes ((:type-prescription)))

(defthm |1^a = 1|
  (equal (o^ 1 a) 1))

(local
 (defthm |(a-1 = 0)  <=>  a <= 1|
   (implies (op a)
	    (equal (equal (o- a 1) 0)
		   (or (equal a 1)
		       (equal a 0))))))

(defthm op-o^
  (implies (and (op a)
                (op b))
           (op (o^ a b)))
  :hints (("goal"
           :induct (o^ a b))))

(verify-guards ob^)

(defthm |(a = 0) /\ ~(b = 0)  <=>  (a^b = 0)|
  (implies (and (op a)
                (op b))
           (equal (equal (o^ a b) 0)
                  (and (equal a 0)
                       (not (equal b 0)))))
  :hints (("goal"
           :induct (o^ a b))))

(defthm o^-^
  (implies (and (natp a)
                (natp b))
           (equal (o^ a b)
                  (expt a b))))

(defthm consp-o^-1
 (implies (and (not (equal b 0))
               (consp a)
	       (op a)
	       (op b))
	  (consp (o^ a b)))
 :hints (("goal"
	  :do-not-induct t
	  :expand (o^ a b)))
 :rule-classes ((:forward-chaining :trigger-terms ((o^ a b)))
		(:rewrite)))

(defthm consp-o^-2
  (implies (and (not (equal a 0))
		(not (equal a 1))
		(consp b)
		(op a)
		(op b))
	   (consp (o^ a b)))
  :hints (("goal"
	   :do-not-induct t
	   :expand (o^ a b)))
  :rule-classes ((:forward-chaining :trigger-terms ((o^ a b)))))

(defthm o^-atom
  (implies (and (not (consp b))
                (not (consp a)))
           (not (consp (o^ a b))))
  :rule-classes :type-prescription)

(defthm consp-o^
  (implies (and (natp a)
                (op b))
           (equal (consp (o^ a b))
                  (and (consp b)
                       (not (equal a 0))
                       (not (equal a 1)))))
  :rule-classes ((:rewrite)
                 (:forward-chaining :trigger-terms ((o^ a b)))))

(encapsulate 

 ; This encapsulate is here to remove a subgoal hint.

 nil
 
 (local
  (defthm o^-first-exp-helper
    (implies (and (consp b)
                  (consp a)
                  (equal (first-exp (o^ a (cdr b)))
                         (o* (first-exp a) (cdr b)))
                  (op a)
                  (op b))
             (equal (o+ (cons (cons (o+ (first-exp (first-exp a))
                                        (first-exp b))
                                    (first-coef b))
                              0)
                        (first-exp (o^ a (cdr b))))
                    (o* (first-exp a) b)))
    :hints (("goal"
             :in-theory (disable |a(b + c) = ab + ac|)
             :use ((:instance |a(b + c) = ab + ac|
                              (a (first-exp a))
                              (b (first-term b))
                              (c (cdr b))))))))

 (defthm o^-first-exp
   (implies (and (consp a)
		 (op a)
                 (op b))
           (equal (first-exp (o^ a b))
                  (o* (first-exp a)
                      b)))
  :hints (("goal"
           :induct (o^ a b)))))
(local
 (defthm |~(a=0)  =>  (a < b  <=>  a-1 < b-1)|
   (implies (and (op a)
		 (op b)
		 (not (equal a 0)))
	    (equal (o< (o- a 1)
		       (o- b 1))
		   (o< a b)))
   :hints (("goal"
	    :in-theory (enable o-)))))

(defthm o^-first-exp-2
  (implies (and (natp a)
		(not (equal a 0))
		(not (equal a 1))
		(op b))
	   (equal (first-exp (first-exp (o^ a b)))
		  (o- (first-exp b) 1))))
  	  	    	  	   
(defthm o^-first-exp-3
  (implies (and (not (equal b 0))
                (consp a)
                (op a)
                (op b))
           (equal (first-exp (first-exp (o^ a b)))
                  (o+ (first-exp (first-exp a)) (first-exp b)))))

(defthm |a < 1  <=> a = 0|
  (implies (op a)
	   (equal (o< a 1)
		  (equal a 0)))
  :rule-classes :forward-chaining)

(local
 (defthm o*-fe-smash-1a
   (implies (and (equal (cdr b) 0)
		 (not (equal a 0))
		 (op a)
		 (op b)
		 (op c)
		 (o< (first-exp (first-exp a))
		     (first-exp (first-exp b))))
	    (equal (o* a b c)
		   (o* b c)))
   :hints (("goal"
	    :in-theory (disable |(ab)c = a(bc)|)
	    :use ((:instance |(ab)c = a(bc)|))))))

(local
 (defthm o*-fe-smash-2a
   (implies (and (atom a)
		 (not (equal a 0))
		 (consp b)
		 (equal (cdr b) 0)
		 (op a)
		 (op b)
		 (op c))
	    (equal (o* a b c) (o* b c)))
   :hints (("goal"
	    :in-theory (disable |(ab)c = a(bc)|)
	    :use ((:instance |(ab)c = a(bc)|))))))

(local
 (defthm |(b > 1)  =>  (a < b  <=>  a-1 < b-1|
   (implies (and (op a)
		 (op b)
		 (not (equal b 1)))
	    (equal (o< (o- a 1)
		       (o- b 1))
		   (o< a b)))
   :hints (("goal"
	    :in-theory (enable o-)))))

(encapsulate

 ;this encapsulate is here to get rid of subgoal hints

 ()

  (local
   (defthm lemma-1
     (implies (and (natp a)
		   (op b)
		   (op c)
		   (o< (first-exp b)
		       (first-exp c)))
	      (equal (o* (o^ a b) (o^ a c))
		     (o^ a c)))
     :hints (("goal"
	      :do-not-induct t
	      :expand (o^ a c)))))

  (local 
   (defthm lemma-2
     (implies (and (consp a)
		   (not (equal b 0)) 
		   (op a)
		   (op b)
		   (op c)
		   (o< (first-exp b)
		       (first-exp c)))
	      (equal (o* (o^ a b) (o^ a c))
		     (o^ a c)))
     :hints (("goal"
	      :do-not-induct t
	      :expand (o^ a c)))))
	     
  (defthm |fe(b) < fe(c)  =>  (a^b)*(a^c) = a^c|
    (implies (and (op a)
		  (op b)
		  (op c)
		  (o< (first-exp b)
		      (first-exp c)))
	     (equal (o* (o^ a b) (o^ a c))
		    (o^ a c)))
    :hints (("goal"
	     :do-not-induct t
	     :cases ((and (not (equal b 0)) 
			  (consp a)))))))

(in-theory (disable o^-first-exp))

(defthm o*-fe-smash-1b
  (implies (and (equal (cdr c) 0)
		(not (equal b 0))
		(op a)
		(op b)
		(op c)
		(op d)
		(o< (first-exp (first-exp b))
		    (first-exp (first-exp c))))
	   (equal (o* (o* a b) (o* c d))
		  (o* (o* a c) d)))
  :hints (("goal"
	   :in-theory (disable o*))))

(defthm o*-fe-smash-2b
  (implies (and (atom b)
		(not (equal b 0))
		(consp c)
		(equal (cdr c) 0)
		(op a)
		(op b)
		(op c)
		(op d))
	   (equal (o* (o* a b) (o* c d))
		  (o*(o* a c) d))))

(encapsulate
 
 ;this encapsulate is here so the lemmas we used to prove |a^(b+c)  =  a^b * a^c|
 ;don't slow down the other proofs in the book.

 ()

 (local 
  (defthm |a^(b+c)  =  a^b * a^c :a|
    (implies (and (op a)
		  (op b)
		  (op c)
		  (o< (first-exp b) 
		      (first-exp c)))
	     (equal (o^ a (o+ b c))
		    (o* (o^ a b) (o^ a c))))))

 (local
  (defthm b-lemma-1
    (implies (and (consp a)
		  (consp b)
		  (op a)
		  (op b))
	     (o< (first-exp (first-exp (o^ a (cdr b))))
		 (o+ (first-exp (first-exp a)) (first-exp b))))
    :hints (("goal"
	     :cases ((equal (cdr b) 0))))))
	    
 (local
  (defthm b-lemma-2
    (implies
     (and (consp b)
	  (op a)
	  (op b)
	  (op c)
	  (equal (first-exp c) (first-exp b))
	  (not (equal a 1))
	  (not (equal a 0))
	  (consp a))
     (equal (o* (cons (cons (cons (cons (o+ (first-exp (first-exp a))
					    (first-exp b))
					(+ (first-coef b) (first-coef c)))
				  0)
			    1)
		      0)
		(o^ a (cdr c)))
	    (o* (o^ a b) (o^ a c))))
    :hints (("goal"
	     :expand ((o^ a b))
	     :use ((:instance |(ab)c = a(bc)|
			      (a (omega-term (omega-term (o+ (first-exp (first-exp a))
							     (first-exp b))
							 (first-coef b))
					     1))
			      (b (omega-term (omega-term (o+ (first-exp (first-exp a))
							     (first-exp b))
							 (first-coef c))
					     1))
			      (c (o^ a (cdr c)))))
	     :do-not-induct t))))

 (local
  (defun exp-ind-1 (a b c)
    (if (atom b)
	(if (not (posp b))
	    (list a b c)
	  (exp-ind-1 a (- b 1) c))
      (list a b c))))

 ;this theorem is here so we don't have to give a subgoal hint to 
 ;|a^(b+c)  =  a^b * a^c :b|
 (local
  (defthm |a^(b+c)  =  a^b * a^c :b helper|
    (implies (and (consp b)
                  (natp a)
                  (op b)
                  (op c)
                  (equal (first-exp c) 1)
                  (not (equal a 1))
                  (not (equal a 0))
                  (equal (first-exp b) 1)
                  (not (posp (cdr b))))
             (equal (o* (cons (cons (+ (first-coef b) (first-coef c))
                                    1)
                              0)
                        (o^ a (cdr c)))
                    (o* (cons (cons (first-coef b) 1) 0)
                        (o* (cons (cons (first-coef c) 1) 0)
                            (o^ a (cdr c))))))
    :hints (("goal"
             :in-theory (disable |fe(b) < fe(c)  =>  (a^b)*(a^c) = a^c|
                                 |(ab)c = a(bc)|)
             :use ((:instance |(ab)c = a(bc)|
                              (a `((,(first-coef b) . 1) . 0))
                              (b `((,(first-coef c) . 1) . 0))
                              (c (o^ a (cdr c))))
                   (:instance |fe(b) < fe(c)  =>  (a^b)*(a^c) = a^c|
                              (b (cdr b))))))))

 (local
  (defthm |a^(b+c)  =  a^b * a^c :b|
    (implies (and (equal (first-exp c) 
			 (first-exp b))
		  (op a)
		  (op b)
		  (op c))
	     (equal (o^ a (o+ b c))
		    (o* (o^ a b) (o^ a c))))
    :hints (("goal"
	     :expand ((o^ a (o+ b c))
		      (o^ a b))
	     :induct (exp-ind-1 a b c)))))
 
 (local
  (defthm |a^(b+c)  =  a^b * a^c :c|
    (implies (and (op a)
		  (op b)
		  (op c)
		  (not (o< (first-exp c) 
			   (first-exp b))))
	     (equal (o^ a (o+ b c))
		    (o* (o^ a b) (o^ a c))))
  :hints (("goal"
	   :cases ((equal (first-exp b) (first-exp c)))))))
   
 (local
  (defun exp-ind-2 (a b c)
    (if (and (consp b)
	     (o< (first-exp c)
		 (first-exp b)))
	(exp-ind-2 a (cdr b) c)
      (list a b c))))

 (defthm |a^(b+c)  =  a^b * a^c|
   (implies (and (op a)
		 (op b)
		 (op c))
	    (equal (o^ a (o+ b c))
		   (o* (o^ a b) (o^ a c))))
   :hints (("goal"
	    :induct (exp-ind-2 a b c)))))

(defthm |(a^b = 1)  <=>  (a = 1) \/ (b = 0)|
  (implies (and (op a)
                (op b))
           (equal (equal (o^ a b) 1)
		  (or (equal a 1)
		      (equal b 0)))))

(defthm o-+
 (implies (and (not (equal a 0))
	       (op a)
	       (op b))
	  (equal (o- (o+ a b) 1)
                 (o+ (o- a 1) b)))
 :hints (("goal"
	  :in-theory (enable o- o+))))

(encapsulate

 ;this encapsulate is here so the lemmas used to prove |a^(bc) = (a^b)^c|
 ;don't slow down the other proofs in the book.

 ()
 (local (in-theory (enable o^-first-exp)))
 
 (local
  (encapsulate

   ;this encapsulate is here because a-lemma-3 causes problems 
   ;with the proof of |a^(bc) = (a^b)^c|
   
   ()
 
   (local
    (defthm a-lemma-1a
      (implies (and (consp b)
		    (not (equal a 0))
		    (not (equal a 1))
		    (consp c)
		    (equal (cdr c) 0)
		    (op a)
		    (op b)
		    (op c))
	       (equal (o^ a (o* b c))
		      (o^ (o^ a b) c)))
      :hints (("goal"
	       :do-not-induct t
	       :in-theory (disable cons-cons-equal-4 
				   consp-o^-2 |a <= b /\ b < c  =>  a < c|)
	       :use ((:instance consp-o^-2)
		     (:instance |a <= b /\ b < c  =>  a < c|
				(a 1)
				(b (first-exp b))
				(c (o+ (first-exp b)
				       (first-exp c)))))))))

   (local
    (defthm a-lemma-1
      (implies (and (not (equal a 0))
		    (not (equal a 1))
		    (consp c)
		    (equal (cdr c) 0)
		    (op a)
		    (op b)
		    (op c))
	       (equal (o^ a (o* b c))
		      (o^ (o^ a b) c)))
      :hints (("goal"
	       :do-not-induct t
	       :cases ((consp b))))))

   (local
    (defthm a-lemma-2
      (implies (and (consp c)
		    (op a)
		    (op b)
		    (op c))
	       (equal (o* (o^ a (o* b (first-term c)))
			  (o^ a (o* b (cdr c))))
		      (o^ a (o* b c))))
      :hints (("goal"
	       :in-theory (disable |a^(b+c)  =  a^b * a^c|)
	       :use ((:instance |a^(b+c)  =  a^b * a^c|
				(b (o* b (first-term c)))
				(c (o* b (cdr c))))
		     (:instance |a(b + c) = ab + ac|
				(a b)
				(b (first-term c))
				(c (cdr c))))))))

   (local
    (defthm a-lemma-3
      (implies (and (not (equal a 0))
		    (not (equal a 1))
		    (consp c)
		    (op a)
		    (op b)
		    (op c))
	       (equal (o^ (o^ a b) c)
		      (o* (o^ (o^ a b) (first-term c))
			  (o^ (o^ a b) (cdr c)))))))

   (defthm |a^(bc) = (a^b)^c :a|
     (implies (and (not (equal (o^ a b) 0))
		   (not (equal (o^ a b) 1))
		   (consp c)
		   (implies (and (op a) (op b) (op (cdr c)))
			    (equal (o^ a (o* b (cdr c)))
				   (o^ (o^ a b) (cdr c))))
		   (op a)
		   (op b)
		   (op c))
	      (equal (o^ a (o* b c))
		     (o^ (o^ a b) c)))
     :hints (("goal"
	      :do-not-induct t
	      :in-theory (disable a-lemma-1 a-lemma-2 a-lemma-3)
	      :use ((:instance a-lemma-1
			       (c (first-term c)))
		    (:instance a-lemma-2)
		    (:instance a-lemma-3)))))))
	
 (local
  (defthm |a^(bc) = (a^b)^c :b|
    (implies (and (not (equal (o^ a b) 0))
		  (not (equal (o^ a b) 1))
		  (not (consp c))
		  (not (zp c))
		  (implies (and (op a) (op b) (op (+ -1 c)))
			   (equal (o^ a (o* b (+ -1 c)))
				  (o^ (o^ a b) (+ -1 c))))
		  (op a)
		  (op b)
		  (op c))
	     (equal (o^ a (o* b c))
		    (o^ (o^ a b) c)))
    :hints (("goal"
	     :do-not-induct t
	     :in-theory (disable |a^(b+c)  =  a^b * a^c|)
	     :use ((:instance |a^(b+c)  =  a^b * a^c|
			      (c (o* b (+ -1 c)))))
	     :expand (o^ (o^ a b) c)))))
    
 (local	   	
  (defun exp-ind-3 (a b c)
    (cond ((equal (o^ a b) 0)
	   (list a b c))
	  ((equal (o^ a b) 1)
	   (list a b c))
	  ((atom c)
	   (if (zp c)
	       (list a b c)
	     (exp-ind-3 a b (- c 1))))
	  (t (exp-ind-3 a b (cdr c))))))

 ;these "helper" lemmas are here to eliminate the need for subgoal hints
 
 (local
  (defthm |a^(bc) = (a^b)^c helper1|
    (implies (and (equal (o^ a b) 1)
                  (op a)
                  (op b)
                  (op c))
             (equal (o^ a (o* b c)) 1))
    :hints (("goal"
             :cases ((equal a 1))))))

 (local 
  (defthm |a^(bc) = (a^b)^c helper2|
    (implies (and (equal (o^ a b) 0)
                  (op a)
                  (op b)
                  (op c)
                  (consp c))
             (equal (o^ a (o* b c)) 0))
    :hints (("goal"
             :cases ((equal a 0))))))

 (local 
  (defthm |a^(bc) = (a^b)^c helper3|
    (implies (and (equal (o^ a b) 0)
                  (op a)
                  (op b)
                  (posp c))
             (equal (o^ a (o* b c)) 0))
    :hints (("goal"
             :cases ((equal a 0))))))

 ;the following 3 theorems seem easy, but for some reason
 ;acl2 needs them to do the proof of |a^(bc) = (a^b)^c|

 (local
  (defthm posp-consp-0
    (implies (and (not (posp a))
		  (op a)
		  (not (consp a)))
	     (equal a 0))
    :rule-classes (:forward-chaining)))

 (local
  (defthm op-o*-fc
   (implies (and (op a)
		 (op b))
	    (op (o* a b)))
   :rule-classes ((:forward-chaining :trigger-terms ((o* a b))))))
 
 (local
  (defthm o*-posp-consp-0
    (implies (and (not (posp (o* a b)))
		  (not (consp (o* a b)))
		  (op a)
		  (op b))
	     (or (equal a 0)
		 (equal b 0)))
    :hints (("goal"
	     :in-theory (disable o*-0)
	     :use ((:instance o*-0))))
    :rule-classes (:forward-chaining)))

 (defthm |a^(bc) = (a^b)^c|
   (implies (and (op a)
		 (op b)
		 (op c))
	    (equal (o^ a (o* b c)) (o^ (o^ a b) c)))
   :hints (("goal"
	    :induct (exp-ind-3 a b c)))))

(defthm |0 < w^a|
  (implies (op a)
	   (o< 0 (o^ (omega) a))))

(defthm w^a-def
 (implies (and (not (equal a 0))
               (op a))
	  (equal (o^ (omega) a)
		(omega-term a 1)))
 :hints (("goal"
	  :induct (o^ (omega) a)
	  :do-not '(eliminate-destructors))))
	  
;(in-theory (disable (:executable-counterpart omega)))

(defthm o^-w-first-exp
 (implies (and (not (equal a 0))
	       (op a)
	       (op b))
	  (equal (o< (first-exp b) (first-exp (o^ (omega) a)))
		 (o< b (o^ (omega) a)))))

;these hints aren't necessary, but speed up the proof
(defthm |b < w^a  =>  b + w^a = w^a|
  (implies (and (op b)
		(op a)
		(o< b (o^ (omega) a)))
	   (equal (o+ b (o^ (omega) a))
		  (o^ (omega) a)))
  :hints (("goal"
	   :do-not-induct t
	   :in-theory (disable o^-w-first-exp)
	   :use o^-w-first-exp)))

(encapsulate
 
 ;this encapsulate is here so the lemmas used to prove |a < b  <=>  c^a < c^b|
 ;don't slow down the rest of the proofs in this book
 
 ()
 
 (local
  (defthm lemma-1
    (implies (and (not (equal c 1))
		  (not (equal c 0))
		  (op a)
		  (op b)
		  (op c)
		  (o< a b))
	     (equal (o^ c b)
		    (o* (o^ c a) (o^ c (o- b a)))))
    :hints (("goal"
	     :do-not-induct t
	     :in-theory (disable |a^(b+c)  =  a^b * a^c|)
	     :use ((:instance |a^(b+c)  =  a^b * a^c|
			      (a c)
			      (b a)
			      (c (o- b a))))))
    :rule-classes ((:rewrite :match-free :all))))
 
 (local
  (defthm |a < b  =>  c^a < c^b|
    (implies (and (not (equal c 1))
		  (not (equal c 0))
		  (op a)
		  (op b)
		  (op c)
		  (o< a b))
	     (o< (o^ c a) (o^ c b)))
    :hints (("goal"
	     :do-not-induct t
	     :in-theory (disable lemma-1 |~(a=0) /\ b>1  <=>  a < ab|)
	     :use ((:instance lemma-1)
		   (:instance |~(a=0) /\ b>1  <=>  a < ab|
			      (a (o^ c a))
			      (b (o^ c (o- b a)))))))))

 (local
  (defthm |a < b  <=>  c^a < c^b helper|
    (implies (and (o< b a)
		  (not (equal c 1))
		  (not (equal c 0))
		  (op a)
		  (op b)
		  (op c))
	     (equal (o< (o^ c a) (o^ c b))
		    (o< a b)))
  :hints (("goal"
	   :in-theory (disable |a < b  =>  c^a < c^b|)
	   :use ((:instance |a < b  =>  c^a < c^b|
			    (a b)
			    (b a)))))))

 (defthm |a < b  <=>  c^a < c^b| ;a < b  =>  c^a < c^b   
   (implies (and (not (equal c 1))
                 (not (equal c 0))
                 (op a)
                 (op b)
                 (op c))
            (equal (o< (o^ c a) (o^ c b))
                   (o< a b)))
   :hints (("goal"
	    :cases ((o< a b) (equal a b) (o< b a)))))
)

;these hints aren't necessary, but they significantly speed up the proof
(defthm |a < b  <=>  w^a + w^b = w^b|
  (implies (and (op a)
		(op b)
		(o< a b))
	   (equal (o+ (o^ (omega) a)
		      (o^ (omega) b))
		  (o^ (omega) b)))
  :hints (("goal"
	   :do-not-induct t
	   :in-theory (disable w^a-def
			       |a < b  <=>  c^a < c^b|
			       |b < w^a  =>  b + w^a = w^a|)
	   :use ((:instance |a < b  <=>  c^a < c^b|
			    (c (omega)))
		 (:instance |b < w^a  =>  b + w^a = w^a|
			    (a b)
			    (b (o^ (omega) a)))))))

;another proof that should be really simple for acl2, but is needed
;for another proof
(local
 (defthm |b-1 <= a+b|
   (implies (and (op a)
		 (op b))
	    (equal (o< (o- b 1)
		       (o+ a b))
		   (not (equal (o- b 1) (o+ a b)))))
   :hints (("goal"
	    :in-theory (enable o- o+)))))

(encapsulate

 ;this encapsulate is to keep the lemmas used to prove |a < b  =>  a^c <= b^c|
 ;from slowing down any other proofs in the book

 ()
 (local	 
  (defthm |a < b  =>  a^c <= b^c :a|
    (implies (and (natp c)
		  (op a)
		  (op b)
		  (o< a b))
	     (equal (o< (o^ a c) (o^ b c))
		    (not (equal (o^ a c) (o^ b c)))))))
		

 (local
  (defthm <-first-exp-<=-2
    (implies (and (consp a)
		  (consp b)
		  (op a)
		  (op b)
		  (o< a b)
		  (not (equal (first-exp (first-exp a))
			      (first-exp (first-exp b)))))
	     (o< (first-exp (first-exp a))
		 (first-exp (first-exp b))))
    :hints (("goal"
	     :in-theory (disable <-first-exp-<=)
	     :use ((:instance <-first-exp-<=
			      (a (first-exp a))
			      (b (first-exp b))))))))

 (local
  (defthm b-lemma-1
    (implies (and (consp c)
		  (consp a)
		  (not (o< (o^ b (cdr c))
			   (o^ a (cdr c))))
		  (op a)
		  (op b)
		  (op c)
		  (o< (first-exp (first-exp a))
		      (first-exp (first-exp b))))
	     (equal (o< (o^ a c) (o^ b c))
		    (not (equal (o^ a c) (o^ b c)))))
    :hints (("goal"
	     :do-not-induct t
	     :in-theory (disable o<-built-in)
	     :use ((:instance o<-built-in
			      (a (o^ a (cdr c)))
			      (b (o^ b (cdr c))))))
	    ("goal'"
	     :cases ((and (o< (o^ a (cdr c))
                              (o^ b (cdr c)))
                          (equal (o+ (first-exp (first-exp a))
                                     (first-exp c))
                                 (o+ (first-exp (first-exp b))
                                     (first-exp c))))
                     (and (not (o< (o^ a (cdr c))
				   (o^ b (cdr c))))
                          (equal (o+ (first-exp (first-exp a))
                                     (first-exp c))
                                 (o+ (first-exp (first-exp b))
                                     (first-exp c)))))))))

 (local
  (defthm |a < b  =>  a^c <= b^c :b|
    (implies (and (consp c)
		  (consp a)
		  (implies (and (op a)
				(op b)
				(op (cdr c))
				(o< a b))
			   (equal (o< (o^ a (cdr c)) (o^ b (cdr c)))
				  (not (equal (o^ a (cdr c))
					      (o^ b (cdr c))))))
		  (op a)
		  (op b)
		  (op c)
		  (o< a b))
	     (equal (o< (o^ a c) (o^ b c))
		    (not (equal (o^ a c) (o^ b c)))))
  :hints (("goal"
	   :do-not-induct t
	   :in-theory (disable b-lemma-1 <-first-exp-<=-2)
	   :use ((:instance b-lemma-1)
		 (:instance <-first-exp-<=-2))))))

 (local
  (defthm |a < b  =>  a^c <= b^c :c|
    (implies (and (consp c)
              (not (consp a))
              (not (equal (first-exp c) 1))
              (implies (and (op a)
                            (op b)
                            (op (cdr c))
                            (o< a b))
                       (equal (o< (o^ a (cdr c)) (o^ b (cdr c)))
                              (not (equal (o^ a (cdr c)) (o^ b (cdr c))))))
              (op a)
              (op b)
              (op c)
              (o< a b))
         (equal (o< (o^ a c) (o^ b c))
                (not (equal (o^ a c) (o^ b c)))))
    :instructions (promote
		   (casesplit (atom b))
		   (casesplit (equal (o- (first-exp c) 1)
				     (o+ (first-exp (first-exp b))
					 (first-exp c))))
		   bash
		   bash
		   (casesplit (equal (o- (first-exp c) 1)
				     (o+ (first-exp (first-exp b))
					 (first-exp c))))
		   bash
		   bash)))

 (local
  (defthm |a < b  =>  a^c <= b^c :d|
    (implies (and (consp c)
		  (not (consp a))
		  (equal (first-exp c) 1)
		  (implies (and (op a)
				(op b)
				(op (cdr c))
				(o< a b))
			   (equal (o< (o^ a (cdr c)) (o^ b (cdr c)))
				  (not (equal (o^ a (cdr c)) (o^ b (cdr c))))))
		  (op a)
		  (op b)
		  (op c)
		  (o< a b))
	     (equal (o< (o^ a c) (o^ b c))
		    (not (equal (o^ a c) (o^ b c)))))
    :instructions(promote
		  (casesplit (equal a 0))
		  bash
		  (casesplit (equal a 1))
		  bash
		  (casesplit (atom b))
		  bash
		  bash)
    :rule-classes ((:rewrite :match-free :all))))

    	     
 (local
  (defun exp-ind-4 (a b c)
    (if (atom c)
	(list a b c)
      (cond ((atom a)
	     (if (equal (first-exp c) 1)
		 (exp-ind-4 a b (cdr c))
	       (exp-ind-4 a b (cdr c))))
	    (t (exp-ind-4 a b (cdr c)))))))

 ;all the subgoals of this proof are proved in the previous lemmas.
 ;all we need to do is get acl2 to use those lemmas instead of trying
 ;to prove the subgoals again. that's why we disable o^, o<, and first-exp-natp.
 (defthm |a < b  =>  a^c <= b^c|
   (implies (and (op a)
		 (op b)
		 (op c)
		 (o< a b))
	    (equal (o< (o^ a c) (o^ b c))
                   (not (equal (o^ a c) (o^ b c)))))
   :hints (("goal"
	    :in-theory (disable o^ o< first-exp-natp)
	    :induct (exp-ind-4 a b c)))))
