#|
This book proves the isomorphism between the acl2 ordinals 
and our cnf ordinals under their respective < relations. the
goal is to represent the high level ideas of bijection and
homomorphism in acl2. To that end, we define ctoa: oc -> ac
(where oc is the set of cnf ordinals and ac is the set of acl2
ordinals) and show it to be an isomorphism in the following manner:

ctoa is well-defined:
That is, we show that x in oc  =>  ctoa.x in oa.
In acl2, this translates in acl2 to the following:

(implies (op x)
	 (e0-ordinalp (ctoa x)))

ctoa is surjective:
That is, y in oa  => there exists x in oc s.t. ctoa.x = y.

To accomplish this, we define atoc: ac -> oc and show that
it is the inverse of ctoa. In acl2, this translates to:

(implies (e0-ordinalp x)
	 (op (ctoa x)))

and

(implies (op x)
	 (equal (atoc (ctoa x))
		x))

and

(implies (e0-ordinalp x)
	 (equal (ctoa (atoc x))
		x))

Hence, ctoa is surjective since for any y in oa, there exists
x = atoc.y such that ctoa.x = ctoa(atoc.y) = y.

ctoa is injective:
That is, x,y in oc  =>  [ctoa.x = ctoa.y  =>  x = y]

This translates in acl2 to

(implies (and (op x)
	      (op y))
	 (equal (equal (ctoa x) (ctoa y))
		(equal x y)))

we also proved the equivalent theorem for atoc

(implies (and (e0-ordinalp x)
	      (e0-ordinalp y))
	 (equal (equal (atoc x) (atoc y))
		(equal x y)))

ctoa is homomorphic with respect to c<:
That is, x,y in oc s.t. x c< y  =>  ctoa.x a< ctoa.y
This translates easily into acl2 as

(implies (and (op x)
	      (op y))
	 (equal (e0-ord-< (ctoa x)
			  (ctoa y))
		(o< x y)))

and equivalently for atoc

(implies (and (e0-ordinalp x)
	      (e0-ordinalp y))
	 (equal (o< (atoc x)
		    (atoc y))
		(e0-ord-< x y)))

Therefore, the above acl2 theorems, which are exported from this file, show
our representation of ordinals in cnf form to be equivalent to acl2's ordinal
representation.

As an added bonus, the above theorems also imply that our o< is a well-founded
relation. Therefore, we can use our representation as acl2's well-founded relation.
At the end of this file we use the set-well-founded-relation command to tell
acl2 to use our relation for termination proofs.
|#

(in-package "ACL2")
(include-book "basic-definitions-thms")
(include-book "ordinal-addition")
(include-book "top-with-meta")

(local (disable-forcing))

(local (in-theory (disable zp-open)))

(local
 (defthm open-zp
   (implies (syntaxp (not (variablep x)))
	    (equal (zp x)
		   (not (posp x))))))

;makes a list of copies of a that is n elements
(defun copyn (a n) 
  (if (zp n) 
      nil 
    (cons a (copyn a (1- n)))))

(local
 (defthm acl2-count-first-exp
   (implies (consp a)
	    (< (acl2-count (first-exp a))
	       (acl2-count a)))
   :hints (("goal"
	    :in-theory (enable first-exp)))))

;converts an ordinal in cnf form to e0-ordinal
(defun ctoa (x)
  (if (atom x)
      x
    (append (copyn (ctoa (first-exp x)) 
		   (first-coef x))
            (ctoa (cdr x)))))

(local
 (defthm |x a<  y  =>  ~(y a< x)|
   (implies (e0-ord-< x y)
	    (not (e0-ord-< y x)))
   :rule-classes :forward-chaining))

(local
 (defthm car-app-copyn
   (implies (posp n)
	    (equal (car (append (copyn x n) y)) x))
   :hints (("goal" :expand (copyn x n)))))

(local
 (defthm posp-consp
   (implies (posp x)
	    (consp (copyn y x)))
   :hints (("goal" :expand (copyn y x)))
   :rule-classes :type-prescription))

(local
 (defthm consp-append
   (implies (consp a)
	    (consp (append a b)))))

(local
 (defthm cnf-cons
   (implies (and (op y)
		 (consp y))
	    (consp (ctoa y)))
   :hints (("goal"
	    :expand (ctoa y)))
   :rule-classes :type-prescription))

(local
 (defthm cnf-l1
   (implies (and (consp x)
		 (consp y)
		 (not (equal (first-exp x) 
			     (first-exp y)))
		 (e0-ord-< (ctoa (first-exp x))
			   (ctoa (first-exp y)))
		 (op x)
		 (op y)
		 (o< (caar x) (caar y)))
	    (e0-ord-< (ctoa x)
		      (ctoa y)))
   :hints (("goal" 
	    :expand ((ctoa x) (ctoa y))))))

; Don't really need the hints above.

(local
 (defun i2 (n1 n2)
   (if (zp n2) 
       n1
     (i2 (- n1 1) (- n2 1)))))

(local
 (defthm copyn-<-2
   (implies (and (<= w v)
		 (natp w)
		 (natp v))
	    (equal (e0-ord-< (append (copyn z w)
				     cx)
			     (append (copyn z v)
				     cy))
		   (e0-ord-< cx (append (copyn z (- v w)) cy))))
   :hints (("goal" :induct (i2 v w)))))

(local
 (defthm acl2-count-first-exp-2
   (implies (atom a)
	    (<= (acl2-count (first-exp a))
		(acl2-count a)))
   :hints (("goal"
	    :in-theory (enable first-exp)))))

(local
 (defthm i1-lemma-1
   (implies (consp a)
	    (< (acl2-count (first-exp (cdr a)))
	       (acl2-count a)))
   :instructions (promote
		  (claim (< (acl2-count (cdr a))
			    (acl2-count a)))
		  (casesplit (atom (cdr a)))
		  bash
		  (claim (< (acl2-count (first-exp (cdr a)))
			    (acl2-count (cdr a))))
		  bash)))

(local
 (defun i1 (x y)
   (if (and (consp x)
	    (consp y))
       (if (equal (first-exp x) (first-exp y))
	   (if (equal (first-coef x) (first-coef y))
	       (i1 (cdr x) (cdr y))
	     (i1 (first-exp (cdr x)) (first-exp y)))
	 (i1 (first-exp x) (first-exp y)))
     (cons x y))))
 
(local
 (defthm app-copyn-1
   (implies (consp (copyn z x))
	    (equal (car (append (copyn z x) y))
		   z))))

(local
(encapsulate

 ;this encapsulate is here so we can eliminate subgoal hints

 ()

 (local
  (defthm o<-e0-ord-<-helper1
    (implies (and (and (consp x) (consp y))
		  (not (equal (first-exp x) (first-exp y)))
		  (implies (and (op (first-exp x))
				(op (first-exp y))
				(o< (first-exp x) (first-exp y)))
			   (e0-ord-< (ctoa (first-exp x))
				     (ctoa (first-exp y)))))
	     (implies (and (op x) (op y) (o< x y))
		      (e0-ord-< (ctoa x)
				(ctoa y))))
    :hints (("goal" :expand ((ctoa x) (ctoa y))))))

 (local
  (defthm o<-e0-ord-<-helper2
    (implies (and (consp y)
		  (equal (first-exp x) (first-exp y))
		  (not (equal (first-coef x) (first-coef y)))
		  (e0-ord-< (ctoa (first-exp (cdr x)))
			    (ctoa (first-exp x)))
		  (op x)
		  (op y)
		  (< (first-coef x) (first-coef y)))
	     (e0-ord-< (append (copyn (ctoa (first-exp x))
				      (first-coef x))
			       (ctoa (cdr x)))
		       (ctoa y)))
    :hints (("goal"
	     :expand ((ctoa y)
		      (ctoa (cdr x)))))))


 (defthm o<-e0-ord-<
   (implies (and (op x)
		 (op y)
		 (o< x y))
	    (e0-ord-< (ctoa x) 
		      (ctoa y)))
   :hints (("goal" :induct (i1 x y))))))

(local
 (encapsulate

  ;this encapsulate is here to eliminate the need for subgoal hints

  ()
  (local
   (defthm helper
     (implies (and (e0-ordinalp (append (copyn z (+ -1 x)) n))
		   (posp x)
		   (natp n)
		   (not (equal z 0))
		   (e0-ordinalp z)
		   (consp (append (copyn z (+ -1 x)) n)))
	      (not (e0-ord-< z
			     (car (append (copyn z (+ -1 x)) n)))))
     :hints (("goal"
	      :cases ((posp (+ -1 x)))))
     :rule-classes ((:rewrite :match-free :all))))
   
  
   (defthm e0-ord-app-copyn
     (implies (and (posp x)
		   (natp n)
		   (not (equal z 0))
		   (e0-ordinalp z))
	      (e0-ordinalp (append (copyn z x) n))))))

(local
 (defun nat-ind (n)
   (if (zp n)
       1
     (cons n (nat-ind (1- n) )))))

(local
 (encapsulate
 
  ;this encapsulate is here to get rid of subgoal hints

  ()

  (local
   (defthm helper
     (implies (and (not (zp n))
		   (e0-ordinalp (append (copyn x (+ -1 n)) y))
		   (not (equal x 0))
		   (e0-ordinalp x)
		   (e0-ordinalp y)
		   (e0-ord-< (car y) x)
		   (consp (append (copyn x (+ -1 n)) y)))
	      (not (e0-ord-< x
			     (car (append (copyn x (+ -1 n)) y)))))
     :hints (("goal"
	      :cases ((zp (+ -1 n)))))
     :rule-classes ((:rewrite :match-free :all))))

  (defthm e0-ordp-app-copy
    (implies (and (not (equal x 0))
		  (e0-ordinalp x)
		  (e0-ordinalp y)
		  (e0-ord-< (car y) x))
	     (e0-ordinalp (append (copyn x n) y)))
    :hints (("goal" :induct (nat-ind n))))))

(local
 (defthm cnf-0
   (implies (op x)
	    (equal (equal (ctoa x) 0)
		   (equal x 0)))))

(local 
 (defthm car-ctoa
   (implies (and (op x)
		 (consp x))
	    (equal (car (ctoa x))
		   (ctoa (first-exp x))))))


(encapsulate

 ()

 (local
  (defthm helper
    (implies (and (consp x)
		  (e0-ordinalp (ctoa (first-exp x)))
		  (e0-ordinalp (ctoa (cdr x)))
		  (op x))
	     (e0-ordinalp (append (copyn (ctoa (first-exp x))
					 (first-coef x))
				  (ctoa (cdr x)))))
    :hints (("goal"
	     :cases ((consp (cdr x)))))))

 (defthm |oc.x  <=>  oa(ctoa.x)|
   (implies (op x) 
	    (e0-ordinalp (ctoa x)))
   :rule-classes ((:forward-chaining)
		  (:rewrite))))

(defun atoc (a)
  (if (atom a)
      a
    (o+ (omega-term (atoc (car a)) 
		    1)
	(atoc (cdr a)))))

(local
 (defthm ateo->=-0
   (implies (and (not (equal a 0))
		 (e0-ordinalp a))
	    (<= 0 (atoc a)))))

(local
 (defthm ateo-neq-0
   (implies (and (not (equal a 0))
		 (e0-ordinalp a))
	    (not (equal 0 (atoc a))))))

(local (in-theory (disable o+)))

(defthm |oa.x  <=>  oc(atoc.x)|
  (implies (e0-ordinalp x)
	   (op (atoc x)))
  :hints (("goal"
	   :induct (atoc x))))

(local
 (defthm consp-atoc
   (implies (and (consp a)
		 (e0-ordinalp a))
	    (consp (atoc a)))
   :instructions (pro
		  (dv 1)
		  x
		  top
		  (claim (not (equal (atoc (car a)) 0)))
		  (claim (op (cons (cons (atoc (car a)) 1) 0)))
		  bash)))
(local
 (defthm e0-total-order
   (implies (and (not (equal a b))
		 (not (e0-ord-< a b))
		 (e0-ordinalp a)
		 (e0-ordinalp b))
	    (e0-ord-< b a))
   :rule-classes :forward-chaining))

(local
 (defun ci (x y)
   (declare (xargs :measure (+ (acl2-count x) (acl2-count y))))
   (if (and (consp x)
	    (consp y))
       (list (ci (cdr x) y)
	     (ci x (cdr y))
	     (ci (cadr x)
		 (car x))
	     (ci (cadr y)
		 (car y))
	     (ci (car x)
		 (car y))
	     (ci (cdr x) 
		 (cdr y))
	     )
     (list x y))))

(local
 (defthm |x a< y  =>  ~(y a< y)|
   (implies (e0-ord-< x y)
	    (not (e0-ord-< y x)))
   :rule-classes :forward-chaining))

(local
 (encapsulate
  
  ()

  (local
   (encapsulate

    ;this encapsulate is here to get rid of the subgoal hints for help1

    ()
  
    (local
     (defthm help1a
       (implies (and (consp x)
		     (consp y)
		     (e0-ordinalp (car x))
		     (not (equal (car x) 0))
		     (e0-ordinalp (cdr x))
		     (not (e0-ord-< (car x) (cadr x)))
		     (e0-ordinalp y)
		     (equal (first-exp (atoc (cdr x)))
			    (atoc (cadr x)))
		     (not (e0-ord-< (cadr x) (car x))))
		(equal (first-exp (o+ (cons (cons (atoc (car x)) 1) 0)
				      (atoc (cdr x))))
		       (atoc (car x))))
       :hints (("goal"
		:use ((:instance e0-total-order
				 (a (cadr x))
				 (b (car x)))
		      (:instance o+-first-exp-2
				 (a (cons (cons (atoc (car x)) 1) 0))
				 (b (atoc (cdr x)))))
		:in-theory (disable e0-total-order o+-first-exp-2)))
       :rule-classes ((:rewrite :match-free :all))))

    (defthm help1
      (implies (and (consp x) 
		    (consp y)  
		    (e0-ordinalp x)
		    (e0-ordinalp y)
		    (implies (and (consp (cdr x))
				  (consp y)
				  (e0-ordinalp (cdr x))
				  (e0-ordinalp y))
			     (equal (first-exp (atoc (cdr x)))
				    (atoc (cadr x))))
		    (implies (and (consp (cadr x))
				  (consp (car x))
				  (e0-ordinalp (cadr x))
				  (e0-ordinalp (car x)))
			     (implies (e0-ord-< (cadr x) (car x))
				      (o< (atoc (cadr x))
					  (atoc (car x))))))
	       (equal (first-exp (atoc x))
		      (atoc (car x))))
      :hints (("goal"
	       :do-not-induct t))
      :rule-classes nil)))

  (local
   (defthm help2
     (implies (and (consp x) 
		   (consp y)  
		   (e0-ordinalp x)
		   (e0-ordinalp y)
		   (equal (first-exp (atoc x))
			  (atoc (car x)))
		   (equal (first-exp (atoc y))
			  (atoc (car y)))
		   (implies (e0-ord-< (car x) (car y))
			    (o< (atoc (car x)) (atoc (car y))))
    
		   (implies (e0-ord-< (cdr x) (cdr y))
			    (o< (atoc (cdr x)) (atoc (cdr y))))
		   (e0-ord-< x y))
	      (o< (atoc x) (atoc y)))
     :rule-classes nil))

  (local
   (defthm lemma
     (implies (and (consp x)
		   (consp y)
		   (e0-ordinalp x)
		   (e0-ordinalp y))
	      (and (equal (first-exp (atoc x))
			  (atoc (car x)))
		   (equal (first-exp (atoc y))
			  (atoc (car y)))
		   (implies (e0-ord-< x y)
			    (o< (atoc x)
				(atoc y)))))
     :rule-classes ((:rewrite :match-free :all))
     :instructions
     ((induct (ci x y))
      (promote)
      (claim (equal (first-exp (atoc x))
		    (atoc (car x)))
	     :hints (("goal"
		      :by help1)))

      (claim (equal (first-exp (atoc y))
		    (atoc (car y)))
	     :hints (("goal"
		      :by (:instance help1 (x y) (y x)))))
      (s)
      (promote)

      (use help2)
      (promote)
      (drop 2 3 4 5 6 7)
      (demote 1)
      (dive 1 1)
      (s-prop)
      (top)
      (casesplit (equal (car x) (car y)))
      (bash)
      (demote 9)
      (dv 1)
      (expand)
      (s)
      (top)
      (promote)
      (dive 1 1 1)
      (s)
      (up)
      (s)
      (top)
      (casesplit (consp (car x)))
      (claim (consp (car y)))
      (bash)
      (drop 1 2)
      (bash))))

  (defthm e0-ord-<-o<
    (implies (and (e0-ordinalp x)
		  (e0-ordinalp y)
		  (e0-ord-< x y))
	     (o< (atoc x)
		 (atoc y)))
    :hints (("goal"
	     :do-not-induct t
	     :use lemma)))

  (defthm first-exp-atoc
    (implies (and (consp x)
		  (e0-ordinalp x))
	     (equal (first-exp (atoc x))
		    (atoc (car x)))))))

(local
 (defthm atoc-not-<
   (implies (and (e0-ordinalp a)
		 (e0-ordinalp b)
		 (not (e0-ord-< a b)))
	    (not (o< (atoc a)
		     (atoc b))))
   :instructions(pro
		 (casesplit (equal a b))
		 (= a b)
		 bash
		 (claim (e0-ord-< b a)
			:hints (("goal" :use e0-total-order)))
		 (claim (o< (atoc b) (atoc a)))
		 bash)))

(local
 (defun ind (x y)
   (if (atom y)
       (cons x y)
     (ind x (cdr y)))))

(local
 (defthm ctao-o+
   (implies (and (op a)
		 (not (equal a 0))
		 (op b)
		 (not (o< a (first-exp b))))
	    (equal (ctoa (o+ (cons (cons a 1) 0) b))
		   (cons (ctoa a) (ctoa b))))
   :instructions(pro
		 (claim (op (cons (cons a 1) 0)))
		 (claim (not (o< (first-exp (cons (cons a 1) 0)) (first-exp b))))
		 (induct (ind a b))
		 pro
		 (casesplit (equal (first-exp b) a))
		 (claim (equal (first-exp (cons (cons a 1) 0)) (first-exp b)))
		 (dv 1)
		 x
		 top
		 bash
		 (claim (o< (first-exp b) (first-exp (cons (cons a 1) 0))))
		 (dv 1)
		 x
		 (dv 1)
		 x
		 top
		 bash
		 bash)))

(defthm |oa.x  =>  ctoa(atoc.x) = x|
  (implies (e0-ordinalp x)
	   (equal (ctoa (atoc x))
		  x))
  :instructions((induct (atoc x))
		(pro)
		(claim (e0-ordinalp (car x)))
		(claim (e0-ordinalp (cdr x)))
		(dv 1 1)
		x
		up
		(rewrite ctao-o+)
		top
		bash
		bash
		bash
		bash
		bash
		(casesplit (atom (cdr x)))
		bash
		(dv 1 2)
		(rewrite first-exp-atoc)
		(claim (not (e0-ord-< (car x) (cadr x))))
		(claim (e0-ordinalp (cadr x)))
		top
		bash
		bash))

(defthm ctoa-=-equiv
  (implies (and (op x)
		(op y))
	   (equal (equal (ctoa x)
			 (ctoa y))
		  (equal x y)))
  :hints (("goal"
	   :do-not-induct t
	   :use ((:instance o<-e0-ord-<
			    (x y)
			    (y x))
		 (:instance o<-e0-ord-<)))))
	  
(defthm atoc-=-equiv
 (implies (and (e0-ordinalp x)
	       (e0-ordinalp y))
	  (equal (equal (atoc x)
			(atoc y))
		 (equal x y)))
 :hints (("goal"
	  :do-not-induct t
	  :use ((:instance e0-ord-<-o<
			   (x y)
			   (y x))
		(:instance e0-ord-<-o<)))))

(defthm |oc.x  =>  atoc(ctoa.x) = x|
  (implies (op x)
	   (equal (atoc (ctoa x))
		  x))
  :hints (("goal"
	   :in-theory (disable ctoa-=-equiv)
	   :use ((:instance ctoa-=-equiv
			    (x (atoc (ctoa x)))
			    (y x))))))

(defthm ctoa-<-equiv
  (implies (and (op x)
		(op y))
	   (equal (e0-ord-< (ctoa x)
			    (ctoa y))
		  (o< x y)))
  :hints (("goal"
	   :use ((:instance e0-ord-<-o<
			    (x (ctoa x))
			    (y (ctoa y)))))))

(defthm atoc-<-equiv
  (implies (and (e0-ordinalp x)
		(e0-ordinalp y))
	   (equal (o< (atoc x)
		      (atoc y))
		  (e0-ord-< x y)))
  :hints (("goal"
	   :use ((:instance o<-e0-ord-<
			    (x (atoc x))
			    (y (atoc y)))))))

(defthm well-founded-cnf
  (and (implies (op x) (e0-ordinalp (ctoa x)))
       (implies (and (op x)                       
                     (op y)
                     (o< x y))
                (e0-ord-< (ctoa x) (ctoa y))))
  :rule-classes :well-founded-relation)

(set-well-founded-relation o<)
