;definitions and theorems about ordinal arithmetic
(in-package "ACL2")
(include-book "ordinal-addition")
(local (disable-forcing))
(local (in-theory (disable o+)))

(defthm o*-atomic
  (implies (and (not (consp a))
		(not (consp b)))
	   (not (consp (o* a b))))
  :rule-classes ((:type-prescription)
                 (:forward-chaining :trigger-terms ((o* a b)))))

(defthm consp-o*
  (implies (and (not (equal b 0))
                (consp a)
                (op a)
                (op b))
           (consp (o* a b)))
  :rule-classes ((:rewrite :match-free :all)
                 (:forward-chaining :trigger-terms ((o* a b)))))

(defthm consp-o*-2
  (implies (and (not (equal b 0))
                (consp a)
                (op a)
                (op b))
           (consp (o* b a)))
  :rule-classes ((:rewrite :match-free :all)
                 (:forward-chaining :trigger-terms ((o* a b)))))

(defthm |a0 = 0|
  (equal (o* a 0) 0))

(defthm |0a = 0|
  (equal (o* 0 a) 0))

(defthm |a1 = a|
  (implies (op a)
           (equal (o* a 1) a)))

(defthm |1a = a|
  (implies (op a)
           (equal (o* 1 a) a)))

(defthm o*-0
 (implies (and (op a)
	       (op b))
	  (equal (equal (o* a b) 0)
                 (or (equal a 0)
                     (equal b 0)))))

(defthm |~(a=0) /\ b>1  <=>  a < ab|
  (implies (and (op a)
                (op b))
           (equal (o< a (o* a b))
                  (and (not (equal a 0))
                       (not (equal b 0))
                       (not (equal b 1)))))
  :hints (("goal"
           :induct (o* a b))))

(defthm |a = ab  <=>  (a = 0) \/ (b = 1)|
  (implies (and (op a)
                (op b))
           (equal (equal a (o* a b))
                  (or (equal a 0) 
                      (equal b 1))))
  :hints (("goal" 
           :cases ((equal a 0)
                   (equal b 0)
                   (equal b 1)))))

(defthm o*-first-exp
 (implies (and (not (equal a 0))
	       (not (equal b 0))
	       (op a)
	       (op b))
	  (equal (first-exp (o* a b))
		 (o+ (first-exp a) (first-exp b))))
 :rule-classes ((:rewrite :match-free :all)))

(encapsulate 
 
 ;this encapsulate is here so that o*-fe-< doesn't interfere with
 ;any other theorems in the book

 ()

 (local
  (defthm o*-fe-<
    (implies (and (op a)
                  (op b)
                  (op c)
                  (o< (first-exp a)
                      (first-exp b)))
             (o< (first-exp (o* c a))
                 (o+ (first-exp c) (first-exp b))))))


 (defthm op-o*
   (implies (and (op a)
                 (op b))
            (op (o* a b)))
   :hints (("goal"
            :in-theory (enable o*-fe-<)
            :induct (o* a b)))))
	    
(defthm |(ab)c = a(bc)|
  (implies (and (op a)
		(op b)
		(op c))
	   (equal (o* (o* a b) c)
		  (o* a b c))))


(defthm o*-first-coef
  (implies (and (not (equal a 0))
		(consp b)
		(op a)
		(op b))
	   (equal (first-coef (o* a b))
		  (first-coef b))))

(defthm o*-first-coef-2
  (implies (and (atom b)
		(op a)
		(op b))
	   (equal (first-coef (o* a b))
		  (* (first-coef a) (first-coef b)))))

(defthm o*-cdr
  (implies (and (not (equal a 0))
		(consp b))
	   (equal (cdr (o* a b))
                  (o* a (cdr b)))))

(defthm o*-cdr-2
  (implies (and (consp a)
		(not (equal b 0))
		(atom b))
	   (equal (cdr (o* a b))
		  (cdr a))))

(defthm o*-first-exp-<
  (implies (and (not (equal a 0))
		(o< (first-exp b)
		    (first-exp c))
		(op a)
		(op b)
		(op c))
	   (o< (first-exp (o* a b))
	       (first-exp (o* a c))))
  :rule-classes ((:forward-chaining 
                  :trigger-terms ((o< (first-exp b)
                                      (first-exp c))))
		 (:rewrite :match-free :all)))

(encapsulate 

; This encapsulate is here to avoid the use of subgoal hints.

 nil

 (local 
  (defthm |a(b + c) = ab + ac helper1|
    (implies (and (o< (first-exp c) (first-exp b))
                  (implies (and (op a) (op (cdr b)) (op c))
                           (equal (o* a (o+ (cdr b) c))
                                  (o+ (o* a (cdr b)) (o* a c))))
                  (op a)
                  (op b)
                  (op c))
             (equal (o* a (o+ b c))
                    (o+ (o* a b) (o* a c))))
    :hints (("goal"
             :in-theory (enable o+)
             :use ((:instance o*-first-exp-<
                              (c b)
                              (b c)))))))

 (local
  (defthm |a(b + c) = ab + ac helper2|
   (implies (and (not (o< (first-exp c) (first-exp b)))
                 (not (o< (first-exp b) (first-exp c)))
                 (op a) 
                 (op b) 
                 (op c))
            (equal (o* a (o+ b c))
                   (o+ (o* a b) (o* a c))))
   :hints (("goal"
            :in-theory (enable o+)))))

 (local
  (defun o*-ind-1 (a b c)
    (cond ((equal a 0)
	   (list a b c))
	  ((and (consp b)
		(o< (first-exp c) 
		    (first-exp b)))
	   (o*-ind-1 a (cdr b) c))
	  ((o< (first-exp b)
	       (first-exp c))
	   (list a b c))
	  (t (list a b c)))))

 (defthm |a(b + c) = ab + ac|
   (implies (and (op a)
                 (op b)
                 (op c))
            (equal (o* a (o+ b c))
                   (o+ (o* a b) (o* a c))))
   :hints (("goal"
            :induct (o*-ind-1 a b c)))))

(defthm o<-o+
  (implies (and (op a)
		(op b)
		(op c)
		(o< a b))
	   (equal (o< (o+ (first-exp a) c)
		      (o+ (first-exp b) c))
		  (not (equal (o+ (first-exp a) c)
			      (o+ (first-exp b) c))))))

(defthm |a < b  =>  ac <= bc|
  (implies (and (op a)
		(op b)
		(op c)
		(o< a b))
	   (equal (o< (o* a c) (o* b c))
		  (not (equal (o* a c)
			      (o* b c))))))

(encapsulate
 
 ;this encapsulate is here so that the lemmas used to prove
 ;|(a < b) /\ ~(c = 0)  <=>  ca < cb| don't slow down the
 ;rest of the book

 ()

 (local
  (defthm |(a < b) /\ ~(c = 0)  <=>  ca < cb :a|
    (implies (and (op a)
                  (op b)
                  (op c)
                  (o< a b)
                  (not (equal c 0)))
             (o< (o* c a) (o* c b)))
    :hints (("goal"
             :induct (o< a b)))))

 (local
  (defthm |(a < b) /\ ~(c = 0)  <=>  ca < cb :b|
    (implies (and (op a)
                  (op b)
                  (op c)
                  (o< (o* c a) (o* c b)))
             (o< a b))
    :hints (("goal"
             :use (:instance |(a < b) /\ ~(c = 0)  <=>  ca < cb :a|  (b a) (a b)))
            ("goal'"
             :cases ((equal a b)
                     (o< a b))))
    :rule-classes ((:rewrite :match-free :all))))

 (defthm |(a < b) /\ ~(c = 0)  <=>  ca < cb| ;a < b  =>  ca < cb    
   (implies (and (op a)
                 (op b)
                 (op c))
            (equal (o< (o* c a) (o* c b))
                   (and (o< a b)
                        (not (equal c 0)))))))

(defthm |ab = ac  <=>  (b=c) \/ (a = 0)|
  (implies (and (op a)
		(op b)
		(op c))
           (equal (equal (o* a b)
                         (o* a c))
                  (or (equal a 0)
                      (equal b c))))
  :hints (("goal"
           :do-not-induct t
           :use ((:instance |(a < b) /\ ~(c = 0)  <=>  ca < cb|
                            (c a)
                            (a b)
                            (b c))
                 (:instance |(a < b) /\ ~(c = 0)  <=>  ca < cb|
                            (c a)
                            (a c)))
           :in-theory (disable |(a < b) /\ ~(c = 0)  <=>  ca < cb|)))
  :rule-classes ((:rewrite :match-free :all)))

(defthm |(a < b) /\ (c < d)  =>  ac < bd|
  (implies (and (op a)
	       (op b)
	       (op c)
	       (op d)
	       (o< a b)
	       (o< c d))
	   (o< (o* a c) (o* b d)))
  :instructions (promote
		 (claim (not (equal b 0)))
		 (casesplit (equal (o* a c) (o* b c)))
		 (= (o* a c) (o* b c))
		 (rewrite |(a < b) / ~(c = 0)  <=>  ca < cb|)
                 s
		 (claim (o< (o* a c) (o* b c)))
		 (claim (o< (o* b c) (o* b d)))
		 bash))

(defthm |(a = 1) /\ (b = 1)  <=>  ab = 1|
  (implies (and (op a)
                (op b))
           (equal (equal (o* a b) 1)
                  (and (equal a 1)
                       (equal b 1))))
  :hints (("goal" 
           :use ((:instance  |~(a=0) /\ b>1  <=>  a < ab| ))
           :in-theory (disable  |~(a=0) /\ b>1  <=>  a < ab| ))))

(defthm o*-1
  (implies (and (op a)
		(posp b))
	   (equal (o+ a (o* a (+ -1 b)))
		  (o* a b)))
  :hints (("goal"
	   :in-theory (enable o+))))

;another seemingly obvious theorem that is needed to help acl2
;reason about equality
(defthm cons-cons-equal-4
  (implies (and (consp a)
		(op a)
                (equal (first-coef a) fca)
                (equal (first-exp a) fea)
                (equal (cdr a) cdra))
           (equal (cons (cons fea fca) cdra)
                  a))
  :rule-classes ((:rewrite :match-free :all)))

(defthm o*-fe-smash-1
  (implies (and (equal (cdr b) 0)
                (not (equal a 0))
                (op a)
                (op b)
                (o< (first-exp (first-exp a))
                    (first-exp (first-exp b))))
           (equal (o* a b)
                  b))
  :rule-classes ((:forward-chaining 
		  :trigger-terms ((o< (first-exp (first-exp a))
				      (first-exp (first-exp b)))))))

(defthm o*-fe-smash-2
  (implies (and (not (equal a 0))
		(consp b)
		(equal (cdr b) 0)
		(natp a)
		(op b))
	   (equal (o* a b) b)))

(defthm o*-atom
  (implies (and (atom a)
		(atom b))
	   (equal (o* a b)
		  (* a b))))

(defthm o*-consp-def
  (implies (not (equal a 0))
           (equal (o* a (cons (cons b1 b2) b3))
                  (cons (cons (o+ (first-exp a) b1) b2)
                        (o* a b3)))))
