(in-package "ACL2")
(include-book "ordinal-definitions")
(local (disable-forcing))

(defthm |~(a<0)|
  (implies (op a)
	   (not (o< a 0)))
  :hints (("goal"
	   :do-not-induct t
	   :cases (atom a))))

(defthm |~(a < a)|
  (not (o< a a)))

(defthm |0 < a  =  ~(a = 0)|
 (implies (op a)
          (equal (o< 0 a)
                 (not (equal a 0)))))

(defthm |~(a < b) /\ ~(a = b)  =>  b < a|
  (implies (and (not (o< a b))
                (not (equal a b))
                (op a)
                (op b))
           (o< b a))
  :hints (("goal"
	   :induct (o< b a)))
  :rule-classes :forward-chaining)

(defthm |b < a  =>  ~(a < b) /\ ~(a = b)|
  (implies (o< b a)
           (and (not (o< a b))
                (not (equal a b))))
  :rule-classes :forward-chaining)

; The above rule is not enough to prove |~(b = 0)  =>  a <= ab|,
; and the next one is needed
(defthm |a < b  =>  ~(a = b)|
 (implies (o< a b)
	  (not (equal a b))))

(defthm |b <= a  =>  (a < b) \/ (a = b)|
  (implies (and (not (o< b a))
                (op a)
                (op b))
           (equal (o< a b)
                  (not (equal a b))))
  :rule-classes :forward-chaining)

(defthm |a < b /\ b < c  =>  a < c|
  (implies (and (o< a b)
                (o< b c))
           (o< a c))
  :rule-classes ((:forward-chaining 
                  :trigger-terms ((o< a b) (o< b c)))
                 (:rewrite :match-free :all)))

(defthm |a < b /\ b <= c  =>  a < c|
  (implies (and (o< a b)
		(op b)
		(op c)
		(not (o< c b)))
	   (o< a c))
  :hints (("goal" :use (:instance |b <= a  =>  (a < b) \/ (a = b)| (b c) (a b))))
  :rule-classes ((:forward-chaining :trigger-terms ((o< a b) (o< b c)))))

(defthm |a <= b /\ b < c  =>  a < c|
  (implies (and (o< b c)
		(op a)
		(op c)
		(not (o< b a)))
	   (o< a c))
  :rule-classes ((:forward-chaining :trigger-terms ((o< b a) (o< b c)))))

(defthm |b <= a /\ a <= b  =>  a = b|
  (implies (and (op a)
                (op b)
                (not (o< a b))
                (not (o< b a)))
           (equal a b))
  :rule-classes ((:forward-chaining
                  :trigger-terms ((o< a b) (o< b a)))))

(defthm |a = b  =>  ~(a < b)|
  (implies (equal a b)
           (not (o< a b)))
  :rule-classes ((:forward-chaining)))

(defthm o<-built-in
  (equal (and (not (o< a b))
              (not (o< b a))
              (op a)
              (op b))
         (and (op a)
              (op b)
              (equal a b)))
  :rule-classes :built-in-clause)

(defthm |(cdr a) < a|
  (implies (and (op a)
		(consp a))
	   (o< (cdr a) a))
  :hints (("goal"
	   :expand ((o< (cdr a) a)))))
